#include "stdio.h"
#include "stdlib.h"
#include "mpi.h"
#include "string.h"
#include <vector>
#include <iostream>
#include <algorithm>
#include "unistd.h"
#include <chrono>

using namespace std;
using namespace std::chrono;

int main(int argc, char **argv)
{
    int rank, size, N;
    MPI_Init(&argc,&argv);
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    MPI_Comm_size(MPI_COMM_WORLD,&size);
    float *Matrix;
    //Read Matrix from file
    if(rank==0)
    {
        FILE *f = fopen(argv[1],"r");
        fscanf(f,"%d",&N);
        MPI_Bcast(&N,1,MPI_INT,0,MPI_COMM_WORLD);
        Matrix = (float*)malloc(N*N*sizeof(float));
        for(int k = 0; k<N*N;k++)
        {
            float fVal;
            fscanf(f,"%f",&fVal);
            Matrix[k]=fVal;
        }
        cout<<endl;
        fclose(f);
    }
    else
    {
        MPI_Bcast(&N,1,MPI_INT,0,MPI_COMM_WORLD);
    }
    int num_rows = N/size;
    float *subMatrix = new float[N*num_rows];
    float *row = new float[N];
    std::chrono::_V2::system_clock::time_point t_start, t_end;
    if(rank==0){
        t_start = (high_resolution_clock::now());
    }
    MPI_Scatter(Matrix,N*num_rows,MPI_FLOAT,subMatrix,N*num_rows,MPI_FLOAT,0,MPI_COMM_WORLD);
    
    //Recieve the Pivot row and do elimination algorithm
    for(int k = 0; k < (rank * num_rows); k++){
        //Recive row
        MPI_Bcast(row,N,MPI_FLOAT,k/num_rows,MPI_COMM_WORLD);
        for(int i = 0; i<num_rows;i++){
            float scale = subMatrix[i * N + k];
            for(int j = k + 1; j < N; j++){
                subMatrix[i * N + j] -= (scale * row[j]);
            }
            subMatrix[i * N + k] = 0;
        }
    }

    int pivot_colum;
    //Eliminate pivot row and do elimination algorithm for rest of the rows
    for(int k = 0; k < num_rows; k++){
        pivot_colum = (rank * num_rows) + k;
        float pivot = subMatrix[k * N + pivot_colum];
        for(int i = pivot_colum + 1; i<N; i++){
            subMatrix[k * N + i] /= pivot;
        }
        subMatrix[k * N + pivot_colum] = 1;
        memcpy(row, &subMatrix[k * N], N * sizeof(float));
        MPI_Bcast(row, N , MPI_FLOAT, rank , MPI_COMM_WORLD);
        for(int i = k + 1; i < num_rows; i++){
            float scale = subMatrix[i * N + pivot_colum];
            for(int j=pivot_colum + 1; j < N; j++){
                subMatrix[i * N + j] -= scale * row[j];
            }
            subMatrix[i * N + pivot_colum] = 0;
        }
    }

    //Wait -> Bcast is synchronous so this for loop is here just to prevent the deadlocks
    for(int k = rank * num_rows + 1; k < N; k++){
        MPI_Bcast(row,N,MPI_FLOAT,k/num_rows,MPI_COMM_WORLD);
    }

    MPI_Gather(subMatrix,N*num_rows,MPI_FLOAT,Matrix,N*num_rows,MPI_FLOAT,0,MPI_COMM_WORLD);
    if(rank==0)
    {
        auto t_end = (high_resolution_clock::now());
        double time = duration_cast<duration<double>>(t_end-t_start).count();
        printf("Total time: %f\n",(time));
        FILE *f = fopen(argv[2],"w");
        fprintf(f,"%d",N);
        for(int k = 0; k<N;k++)
        {
            if((k%N)==0)
                fprintf(f,"\n");
            fprintf(f,"%0.3f ",Matrix[k]);
        }
        fclose(f);
    }
    MPI_Finalize();
    return 0;
}