#include "stdio.h"
#include "stdlib.h"
#include "mpi.h"
#include "string.h"
#include <vector>
#include <iostream>
#include <algorithm>
#include "unistd.h"
#include <chrono>

using namespace std;
using namespace std::chrono;

int main(int argc, char **argv)
{
    int rank, size, N;
    MPI_Init(&argc,&argv);
    MPI_Comm_rank(MPI_COMM_WORLD,&rank);
    MPI_Comm_size(MPI_COMM_WORLD,&size);
    float *Matrix;
    //Read Matrix from file
    if(rank==0){
        FILE *f = fopen(argv[1],"r");
        fscanf(f,"%d",&N);
        MPI_Bcast(&N,1,MPI_INT,0,MPI_COMM_WORLD);
        Matrix = (float*)malloc(N*N*sizeof(float));
        for(int k = 0; k<N*N;k++){
            float fVal;
            fscanf(f,"%f",&fVal);
            Matrix[k]=fVal;
        }
        fclose(f);
    }
    else
    {
        MPI_Bcast(&N,1,MPI_INT,0,MPI_COMM_WORLD);
    }
    int num_rows = N/size;
    float *subMatrix = new float[N * num_rows];
    float *row = new float[N];
    std::chrono::_V2::system_clock::time_point t_start, t_end;
    if(rank==0){
        t_start = (high_resolution_clock::now());
    }
    if(size==1){
        memcpy(subMatrix,Matrix,N*N*sizeof(float));
    }
    else{
        for(int k = 0; k < num_rows;k++){
            MPI_Scatter(&Matrix[k * N * size], N, MPI_FLOAT,
                &subMatrix[k * N], N, MPI_FLOAT, 0, MPI_COMM_WORLD);
        }
    }
    int local_row, pivot_rank;
    float pivot, scale;
    //Iteration is through all rows
    for(int k = 0; k < N; k++){
        local_row = k/size;
        pivot_rank = (k%size);

        if(rank == pivot_rank){
            pivot = subMatrix[local_row * N + k];
            for(int i = k + 1; i < N; i++){
                subMatrix[local_row * N + i] /= pivot;
            }

            subMatrix[local_row * N + k] = 1;

            memcpy(row,&subMatrix[local_row * N], N * sizeof(float));

            MPI_Bcast(row,N,MPI_FLOAT,pivot_rank,MPI_COMM_WORLD);

            for(int i = local_row + 1; i<num_rows;i++){
                scale = subMatrix[i * N + k];
                for(int j = k + 1; j < N; j++){
                    subMatrix[i * N + j] -= scale * row[j];
                }
                subMatrix[i * N + k] = 0;
            }
        }
        else{
            MPI_Bcast(row,N,MPI_FLOAT,pivot_rank,MPI_COMM_WORLD);
            for(int i = local_row; i < num_rows; i++){
                if((pivot_rank < rank) || (i > local_row)){
                    scale = subMatrix[i * N + k];
                    for(int j = k + 1; j < N; j++){
                        subMatrix[i * N + j] -= scale * row[j];
                    }
                    subMatrix[i * N + k] = 0;
                }
            }
        }
    }
    MPI_Barrier(MPI_COMM_WORLD);
    if(size==0){
        memcpy(Matrix,subMatrix,N*N*sizeof(float));
    }
    else{
        for(int k = 0; k < num_rows; k++){
            MPI_Gather(&subMatrix[k * N], N, MPI_FLOAT, &Matrix[k * size * N], N, MPI_FLOAT, 0, MPI_COMM_WORLD);
        }
    }

    if(rank==0){
        auto t_end = (high_resolution_clock::now());
        double time = duration_cast<duration<double>>(t_end-t_start).count();
        printf("Total time: %f\n",(time));
        FILE *f = fopen(argv[2],"w");
        fprintf(f,"%d",N);
        for(int k = 0; k<N*N;k++)
        {
            if((k%N)==0)
                fprintf(f,"\n");
            fprintf(f,"%0.3f ",Matrix[k]);
        }
        fclose(f);
    }
    MPI_Finalize();
    return 0;
}