#include <iostream>
#include <iomanip>
#include <cmath>
#include <omp.h>
#include <time.h>
#include <stdio.h>
#include <stdlib.h>  
#include <chrono>

#define VERBOSE true
#define VALIDATION true
#define DEBUG true

using namespace std;
using namespace std::chrono;

//initialize the matrices
inline void initialize_matrices(float** a, float** l, float** u, int size)
{
	#pragma omp for schedule(static)
	for (int i = 0; i < size; ++i)
	{
		a[i] = new float[size];
		l[i] = new float[size];
		u[i] = new float[size];
	}
}

bool matrix_validation(float** a, float** l, float** u, int size)
{
	float** check = new float* [size];
	float** a2 = new float* [size];
	float** l2 = new float* [size];
	float** u2 = new float* [size];
	initialize_matrices(a2, l2, u2, size);
	for (int i = 0; i < size; ++i)
	{
		check[i] = new float[size];
	}

	a2 = a;
	l2 = l;
	u2 = u;

	//matrix multiplication
	for (int i = 0; i < size; i++)
	{
		for (int j = 0; j < size; j++)
		{
			check[i][j] = 0;
			for (int k = 0; k < size; k++)
			{
				check[i][j] += u2[k][j] * l2[i][k];
			}
		}
	}
	int error = 0;
	cout << "Validation Results:  " << endl;
	for (int i = 0; i < size; i++)
	{
		for (int j = 0; j < size; j++)
		{
			if (abs(check[i][j] != a[i][j])>1e8)
			{
				error = 1;
				cout << "Error index is : " << (check[i][j]-a[i][j]) << endl;
			}

		}
	}

	if (error == 1)
	{
		cout << "Failed" << endl;
		return false;
	}
	else
	{
		cout << "Success" << endl;
		return true;
	}
}

int main(int argc, char** argv)
{
	double runtime;
	int size, chunkSize;
	// user input section end
	#if defined(_OPENMP)
	omp_set_num_threads(16);
	#endif
	FILE *f = fopen(argv[1],"r");
    fscanf(f,"%d",&size);
	float** a = new float* [size];
	float** l = new float* [size];
	float** u = new float* [size];
    initialize_matrices(a, l, u, size);
    for(int k = 0; k<size;k++)
    {
		for(int r = 0; r<size;r++)
		{
			float fVal;
			fscanf(f,"%f",&fVal);
			a[k][r]=fVal;
		}
    }
    fclose(f);
	chunkSize = 1;
	//do LU decomposition
	#if defined(_OPENMP)
	omp_set_num_threads(16);
	#endif
	auto t_start = (high_resolution_clock::now());

	#pragma omp parallel shared(a,l,u)
	{
		for (int i = 0; i < size; i++)
		{
			#pragma omp for schedule(dynamic, chunkSize)
			for (int j = 0; j < size; j++)
			{
				if (j < i){
					l[j][i] = 0;
					continue;
				}
				if (j == i){
					l[i][j] = 1;
					continue;
				}
				float tmp_reduction = a[j][i];
				#pragma omp simd reduction(+:tmp_reduction)
				for (int k = 0; k < i; k++){
					tmp_reduction = tmp_reduction - l[j][k] * u[k][i];
				}
				l[j][i] = tmp_reduction;
			}
			#pragma omp for schedule(dynamic, chunkSize)
			for (int j = 0; j < size; j++)
			{
				if (j < i){
					u[i][j] = 0;
					continue;
				}
				float tmp_reduction = a[i][j] / l[i][i];
				#pragma omp simd reduction(+:tmp_reduction)
				for (int k = 0; k < i; k++)
				{
					tmp_reduction = tmp_reduction - ((l[i][k] * u[k][j]) / l[i][i]);
				}
				u[i][j] = tmp_reduction;
			}
		}
	}
	auto t_end = (high_resolution_clock::now());
    double time = duration_cast<duration<double>>(t_end-t_start).count();
	cout << "LUD Decomposition Time: ";
	cout << time;
	cout << " seconds" << endl;
	if(matrix_validation(a,l,u,size))
	{
		FILE *g = fopen(argv[2],"w");
		fprintf(g,"%d\n",size);
		for(int r = 0; r<size;r++)
		{
			for(int c = 0; c<size;c++){
				fprintf(g,"%0.3f ",l[r][c]);
			}
			fprintf(g,"\n");
		}
		fclose(g);
		FILE *h = fopen(argv[3],"w");
		fprintf(g,"%d\n",size);
		for(int r = 0; r<size;r++)
		{
			for(int c = 0; c<size;c++){
				fprintf(h,"%0.3f ",u[r][c]);
			}
			fprintf(h,"\n");
		}
		fclose(h);
	}
	return 0;
}




