#include "stdio.h"
#include "stdlib.h"
#include "iostream"
#include "omp.h"
#include <vector>
#include "math.h"
#include <chrono>


using namespace std;
using namespace std::chrono;

vector<vector<double>> MatMul(vector<vector<double>> &A, vector<vector<double>> &B)
{
    // cout<<"A shape : <"<<A.size()<<","<<A[0].size()<<">"<<endl;
    // cout<<"B shape : <"<<B.size()<<","<<B[0].size()<<">"<<endl;
    vector<vector<double>> C(A.size(),vector<double>(B[0].size(),0.0));
    for(int i = 0; i < A.size(); i++)
    {
        for(int j = 0; j < B[0].size(); j++)
        {
            double fVal = 0;
            #pragma omp simd reduction(+:fVal)
            for(int k = 0; k<A[0].size(); k++){
                fVal += A[i][k] * B[k][j];
            }
            C[i][j] = fVal;
        }
    }
    return C;
}

vector<vector<double>> MatSum(vector<vector<double>> &A, vector<vector<double>> &B)
{
    vector<vector<double>> C(A.size(),vector<double>(B[0].size(),0.0));
    for(int i = 0; i<A.size(); i++)
    {
        #pragma omp simd
        for(int j = 0; j<A[0].size(); j++)
        {
            C[i][j] = A[i][j] + B[i][j];
        }
    }
    return C;
}

void Load_Matrix(int iRows, int iCols, vector<vector<double>> &A, FILE *f)
{
    for(int i = 0; i<iRows;i++)
    {
        vector<double> tmp_row;
        for(int j = 0; j<iCols; j++)
        {
            double fVal;
            fscanf(f,"%f",&fVal);
            tmp_row.push_back(fVal);
        }
        A.push_back(tmp_row);
    }
}

int main(int argc, char** argv)
{

    vector<vector<double>> A, B, C, D, E, F, G, H, K, L, M, N, Res;
    FILE *f = fopen(argv[1],"r");
    Load_Matrix(2000,2000,A,f);
    fclose(f);
    f = fopen(argv[2],"r");
    Load_Matrix(2000,500,B,f);
    fclose(f);
    f = fopen(argv[3],"r");
    Load_Matrix(500,2000,C,f);
    fclose(f);
    f = fopen(argv[4],"r");
    Load_Matrix(500,2000,D,f);
    fclose(f);
    f = fopen(argv[5],"r");
    Load_Matrix(2000,500,E,f);
    fclose(f);
    f = fopen(argv[6],"r");
    Load_Matrix(500,500,F,f);
    fclose(f);
    int d, e, g;
    auto t_start = (high_resolution_clock::now());
    #pragma omp parallel
    {
        #pragma omp single
        {
            #pragma omp task shared(A , B, K) depend(out: K)
            {
                K = MatMul(A,B);
            }
            #pragma omp task shared(C , A, L) depend(out: L)
            {
                L = MatMul(C,A);
            }
            #pragma omp task shared(E , D, M) depend(out: M)
            {
                M = MatMul(E,D);
            }
            #pragma omp task shared(K , L, N) depend(in: K,L) depend(out: N)
            {
                N = MatMul(K,L);
            }
            #pragma omp task shared(M , F, G) depend(in: M) depend(out: G)
            {
                G = MatSum(M,F);
            }
            #pragma omp task shared(N, G, H) depend(in: N, G) depend(out: H)
            {
                H = MatSum(N,G);
            }

        }
    }
    auto t_end = (high_resolution_clock::now());
    double time = duration_cast<duration<double>>(t_end-t_start).count();
	cout << "Time: ";
	cout << time;
	cout << " seconds" << endl;
    f = fopen(argv[5],"r");
    Load_Matrix(500,500,Res,f);
    fclose(f);
    bool Correct = true;
    for(int k = 0; (k<Res.size());k++)
    {
        for(int l = 0; l<Res[0].size();l++)
        {
            if(abs(H[k][l] - Res[k][l])>0.0001)
            {
                Correct = false;
                cout<<"Error "<<k <<" "<<l<<" "<<abs(H[k][l] - Res[k][l])<<endl;
            }
        }
    }
    if(Correct)
        cout<<"Succes"<<endl;
    return 0;
}